"""
Unit and regression test for the starling package.
"""

# Import package, test suite, and other packages as needed
import sys
import os
import numpy as np

import pytest

import starling

from starling import generate, load_ensemble
from starling.structure.ensemble import Ensemble
from starling import utilities

from soursop.sstrajectory import SSTrajectory

import torch


def test_starling_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "starling" in sys.modules



def test_ensemble_generation_save_and_load():

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    
    E = generate(seq,conformations=10, verbose=False, show_progress_bar=False, return_data=True, return_structures=False, return_single_ensemble=True)
    assert len(E) == 10

    # check we can write a trajectory and load it uncompressed
    E.save('outdata/test_uncompressed')
    E_test_uncompressed = load_ensemble('outdata/test_uncompressed.starling')
    assert np.all(E_test_uncompressed.end_to_end_distance() == E.end_to_end_distance())
    os.remove('outdata/test_uncompressed.starling')

    ##
    ## COMPRESSION TESTS
    ## 

    # check we can write a trajectory using default compression and load 
    E.save('outdata/test_compressed', compress=True)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default lzma explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='lzma')
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default gzip explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='gzip')
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.gzip')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.gzip')    


    ## now check we can do compresion with reduce precision explicitly to False (means should be lossless)

    # check we can write a trajectory using default compression and load 
    E.save('outdata/test_compressed', compress=True, reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(E_test_compressed.end_to_end_distance() == E.end_to_end_distance())
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default lzma explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='lzma',  reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(E_test_compressed.end_to_end_distance() == E.end_to_end_distance())
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default gzip explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='gzip',  reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.gzip')
    assert np.all(E_test_compressed.end_to_end_distance() == E.end_to_end_distance())
    os.remove('outdata/test_compressed.starling.gzip')    

    ## now check we can do compresion with reduce precision explicitly to True (again should be lossy)

    # check we can write a trajectory using default compression and load 
    E.save('outdata/test_compressed', compress=True, reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default lzma explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='lzma',  reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.xz')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.xz')

    # check we can write a trajectory using default gzip explicitly
    E.save('outdata/test_compressed', compress=True, compression_algorithm='gzip',  reduce_precision=False)
    E_test_compressed = load_ensemble('outdata/test_compressed.starling.gzip')
    assert np.all(np.isclose(E_test_compressed.end_to_end_distance(), E.end_to_end_distance(), atol=0.1))
    os.remove('outdata/test_compressed.starling.gzip')    




def test_ensemble_generation():

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False)
    E = C['sequence_1']

    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a trajectory
    t = E.trajectory
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)


def test_ensemble_generation():

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False)
    E = C['sequence_1']

    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a 
    t = E.trajectory
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)

    # check we can write a trajectory
    E.save_trajectory('outdata/test.pdb', pdb_trajectory=True)


def test_ensemble_generation_single_ensemble():

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    # check we can get a single Ensembe object    
    E = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False, return_single_ensemble=True)
    
    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a 
    t = E.trajectory
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)

    # check we can write a trajectory
    E.save_trajectory('outdata/test.pdb', pdb_trajectory=True)

def test_ensemble_generation_cpu():

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False, device='cpu')
    E = C['sequence_1']

    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a 
    t = E.build_ensemble_trajectory(device='cpu')
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)

    # check we can write a trajectory
    E.save_trajectory('outdata/test.pdb', pdb_trajectory=True)



def test_ensemble_generation_mps():

    if not torch.backends.mps.is_available():
        pytest.skip("MPS not available")

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False, device='mps')
    
    E = C['sequence_1']

    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a 
    t = E.build_ensemble_trajectory(device='mps')
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)

    # check we can write a trajectory
    E.save_trajectory('outdata/test.pdb', pdb_trajectory=True)


def test_ensemble_generation_cuda():

    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")
        

    # define sequence
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    
    
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=False, device='cuda')
            
    E = C['sequence_1']

    assert len(E) == 100
    assert abs(E.radius_of_gyration(return_mean=True) - 32) < 3
    assert abs(E.end_to_end_distance(return_mean=True) - 85) < 8
    assert E.sequence == seq

    # check we can build a 
    t = E.build_ensemble_trajectory(device='cuda')
    np.isclose(np.mean(t.get_radius_of_gyration()) , E.radius_of_gyration(return_mean=True), rtol=0.01, atol=0.01)

    # check we can write a trajectory
    E.save_trajectory('outdata/test.pdb', pdb_trajectory=True)
    
def test_ensemble_reconstruction_re():
    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    C = generate(seq,conformations=100, verbose=False, show_progress_bar=False, return_data=True, return_structures=True)
    
    E = C['sequence_1']
    p = E.trajectory
    
    # absolute tollerance of 10 Angstroms
    assert np.all(np.isclose(p.get_end_to_end_distance(), E.end_to_end_distance(), atol=10, rtol=0))

def test_ensemble_reconstruction_dm():
    #
    # MDS reconstruction is hard, so our tolerance here is ~5% of frames can have ONE OR MORE distance that 
    # is off by more than 10 Angstroms. This translates to a very low average error, but we do want to do
    # the full comparison to get a sense of the 'true' error here...
    #

    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    n_confs = 100
    E = generate(seq,conformations=n_confs, verbose=False, show_progress_bar=False, return_single_ensemble=True, return_structures=True)

    P = E.trajectory

    # all structural distance maps ("reconstruction")
    A = utilities.symmetrize_distance_maps(P.get_distance_map(return_instantaneous_maps=True, verbose=False)[0])

    # all STARLING distance maps ("truth")
    B = E.distance_maps()

    # the code below finds the number of frames where 1 or more i-j distances are off by more than 10 angstroms
    mask = np.isclose(A, B, atol=10, rtol=0)  # Boolean array
    offending_indices = np.where(~mask)  # Indices where values are NOT clos
    bad_frames = len(set(offending_indices[0]))

    # 5% of frames are allowed to be bad
    assert bad_frames <= (1+int(n_confs)*0.05)  
    

def test_ensemble_reconstruction_dm_CPU():
    #
    # MDS reconstruction is hard, so our tolerance here is ~10% of frames can have ONE OR MORE distance that 
    # is off by more than 10 Angstroms. This translates to a very low average error, but we do want to do
    # the full comparison to get a sense of the 'true' error here...
    #

    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    n_confs = 100
    E = generate(seq,conformations=n_confs, verbose=False, show_progress_bar=False, return_single_ensemble=True)

    # specifically build with CPU regardless of what was used for generation...
    E.build_ensemble_trajectory(device='cpu')

    P = E.trajectory

    # all structural distance maps ("reconstruction")
    A = utilities.symmetrize_distance_maps(P.get_distance_map(return_instantaneous_maps=True, verbose=False)[0])

    # all STARLING distance maps ("truth")
    B = E.distance_maps()

    # the code below finds the number of frames where 1 or more i-j distances are off by more than 10 angstroms
    mask = np.isclose(A, B, atol=10, rtol=0)  # Boolean array
    offending_indices = np.where(~mask)  # Indices where values are NOT clos
    bad_frames = len(set(offending_indices[0]))

    # ~10% of frames are allowed to be bad
    assert bad_frames <= (1+int(n_confs)*0.10)  


def test_ensemble_reconstruction_dm_mps():
    #
    # MDS reconstruction is hard, so our tolerance here is ~10% of frames can have ONE OR MORE distance that 
    # is off by more than 10 Angstroms. This translates to a very low average error, but we do want to do
    # the full comparison to get a sense of the 'true' error here...
    #
    if not torch.backends.mps.is_available():
        pytest.skip("MPS not available")

    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    n_confs = 100
    E = generate(seq,conformations=n_confs, verbose=False, show_progress_bar=False, return_single_ensemble=True)

    # specifically build with CPU regardless of what was used for generation...
    E.build_ensemble_trajectory(device='mps')

    P = E.trajectory

    # all structural distance maps ("reconstruction")
    A = utilities.symmetrize_distance_maps(P.get_distance_map(return_instantaneous_maps=True, verbose=False)[0])

    # all STARLING distance maps ("truth")
    B = E.distance_maps()

    # the code below finds the number of frames where 1 or more i-j distances are off by more than 10 angstroms
    mask = np.isclose(A, B, atol=10, rtol=0)  # Boolean array
    offending_indices = np.where(~mask)  # Indices where values are NOT clos
    bad_frames = len(set(offending_indices[0]))

    # ~10% of frames are allowed to be bad
    assert bad_frames <= (1+int(n_confs)*0.10)  

def test_ensemble_reconstruction_dm_cuda():
    #
    # MDS reconstruction is hard, so our tolerance here is ~10% of frames can have ONE OR MORE distance that 
    # is off by more than 10 Angstroms. This translates to a very low average error, but we do want to do
    # the full comparison to get a sense of the 'true' error here...
    #
    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")


    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'
    n_confs = 100
    E = generate(seq,conformations=n_confs, verbose=False, show_progress_bar=False, return_single_ensemble=True)

    # specifically build with CPU regardless of what was used for generation...
    E.build_ensemble_trajectory(device='cuda')

    P = E.trajectory

    # all structural distance maps ("reconstruction")
    A = utilities.symmetrize_distance_maps(P.get_distance_map(return_instantaneous_maps=True, verbose=False)[0])

    # all STARLING distance maps ("truth")
    B = E.distance_maps()

    # the code below finds the number of frames where 1 or more i-j distances are off by more than 10 angstroms
    mask = np.isclose(A, B, atol=10, rtol=0)  # Boolean array
    offending_indices = np.where(~mask)  # Indices where values are NOT clos
    bad_frames = len(set(offending_indices[0]))

    # ~10% of frames are allowed to be bad
    assert bad_frames <= (1+int(n_confs)*0.10)  


def test_skip_long_seqs():
    """
    Check we can pass a sequence that's too long and it's skipped but
    does not trigger a total failure.
    """
    
    seqs = {}
    seqs['a'] = 'AP'*20
    seqs['b'] = 'AP'*200

    C = generate(seqs, conformations=10, verbose=False, show_progress_bar=False, return_data=True, return_structures=True)
    assert len(C) == 1

    
def test_invalid_input_options():

    seq = 'ASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAPSPASPASPAPSPASPAPSPPASPASPAASAPASPAPSPAP'

    seqs = {}
    seqs['a'] = 'AP'*20
    seqs['b'] = 'AQ'*30


    # check we fail if return_data=False and output_directory is None
    with pytest.raises(ValueError):
        generate(seq, conformations=10, verbose=False, show_progress_bar=False, return_data=False)
    
    # check we fail if we pass multiple sequences and request a single ensemble
    with pytest.raises(ValueError):
        generate(seqs, conformations=10, verbose=False, show_progress_bar=False, return_single_ensemble=True)


    



import yaml


def get_params(config_file: str = None) -> dict:
    """
    A function that reads the default configuration file
    and merges it with the user configuration file.

    Parameters
    ----------
    config_file : str, optional
        A path to the user configuration file, by default None

    Returns
    -------
    dict
        A dictionary containing the configuration parameters
    """
    with open(config_file, "r") as stream:
        user_config = yaml.safe_load(stream)

    return user_config


def get_vae_params(config_file: str) -> dict:
    """
    A function that reads the default configuration file
    and merges it with the user configuration file.

    Parameters
    ----------
    config_file : str, optional
        A path to the user configuration file, by default None

    Returns
    -------
    dict
        A dictionary containing the configuration parameters
    """

    with open(config_file, "r") as stream:
        user_config = yaml.safe_load(stream)

    return user_config



import argparse
import glob
import os

import pytorch_lightning as pl
import wandb
import yaml
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger
from pytorch_lightning.utilities.rank_zero import rank_zero_only

from starling.data.argument_parser import get_params
from starling.data.ddpm_loader import MatrixDataModule
from starling.models.diffusion import DiffusionModel
from starling.models.unet import UNetConditional
from starling.models.vae import VAE


@rank_zero_only
def wandb_init(project: str = "starling-vista-diffusion"):
    wandb.init(project=project)


def train_model():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Path to the configuration file to use",
    )

    parser.add_argument(
        "--num_nodes",
        type=int,
        default=1,
        help="Number of nodes to use for training",
    )

    parser.add_argument(
        "--num_workers",
        type=int,
        default=16,
        help="Number of workers to use for loading in the data",
    )

    args = parser.parse_args()

    config = get_params(config_file=args.config_file)

    # Make the directories to save the model and logs
    os.makedirs(config["training"]["output_path"], exist_ok=True)
    with open(f"{config['training']['output_path']}/config.yaml", "w") as f:
        yaml.dump(config, f)

    wandb_init(config["training"]["project_name"])

    # Set up model checkpoint saving
    checkpoint_callback = ModelCheckpoint(
        monitor="epoch_val_loss",  # Monitor validation loss for saving the best model
        dirpath=f"{config['training']['output_path']}/",  # Directory to save checkpoints
        filename="model-kernel-{epoch:02d}-{epoch_val_loss:.2f}",  # File name format for saved models
        save_top_k=1,
        mode="min",  # Minimize the monitored metric (val_loss)
    )

    # delta = datetime.timedelta(hours=1)
    save_last_checkpoint = ModelCheckpoint(
        dirpath=f"{config['training']['output_path']}/",  # Directory to save checkpoints
        filename="last",
    )

    checkpoint_dir = f"{config['training']['output_path']}/"
    checkpoint_pattern = os.path.join(checkpoint_dir, "last.ckpt")

    # Check if any checkpoint exists
    checkpoint_files = glob.glob(checkpoint_pattern)

    if checkpoint_files:
        # This will load the most recent checkpoint
        ckpt_path = "last"
    else:
        # This will start training from scratch
        ckpt_path = None

    lr_monitor = LearningRateMonitor(logging_interval="step")

    # Set up data loaders
    dataset = MatrixDataModule(
        **config["data"],
        labels=config["diffusion"]["labels"],
        num_workers=args.num_workers,
    )

    dataset.setup(stage="fit")
    encoder_model_path = config["unet"].pop("encoder_path")
    UNet_model = UNetConditional(**config["unet"])

    map_location = "cuda:0"
    encoder_model = VAE.load_from_checkpoint(
        encoder_model_path, map_location=map_location
    )

    # TODO
    # SHOULD CHANGE TO DYNAMICALLY AUTOMATICALLY GET LATEN_DIM IN DIFFUSION.PY
    diffusion_model = DiffusionModel(
        model=UNet_model,
        encoder_model=encoder_model,
        **config["diffusion"],
    )

    with open(f"{config['training']['output_path']}/model_architecture.txt", "w") as f:
        f.write(str(UNet_model))

    # Set up logging on weights and biases
    wandb_logger = WandbLogger(project=config["training"]["project_name"])
    wandb_logger.watch(diffusion_model)

    # Set up PyTorch Lightning Trainer
    trainer = pl.Trainer(
        accelerator="auto",
        devices=config["device"]["cuda"],
        num_nodes=args.num_nodes,
        max_epochs=config["training"]["num_epochs"],
        callbacks=[checkpoint_callback, lr_monitor, save_last_checkpoint],
        gradient_clip_val=config["training"]["gradient_clip_val"],
        precision="bf16-mixed",
        logger=wandb_logger,
    )

    # Start training
    trainer.fit(diffusion_model, dataset, ckpt_path=ckpt_path)

    # Detach the logging on wandb
    wandb_logger.experiment.unwatch(diffusion_model)
    wandb.finish()


if __name__ == "__main__":
    train_model()



import importlib.util
import os

from starling.utilities import fix_ref_to_home

# stand-alone default parameters
# NB: you can overwrite these by adding a configs.py file to ~/.starling_weights/
DEFAULT_MODEL_DIR = os.path.join(
    os.path.expanduser(os.path.join("~/", ".starling_weights"))
)
DEFAULT_ENCODE_WEIGHTS = "model-kernel-epoch=99-epoch_val_loss=1.72.ckpt"
DEFAULT_DDPM_WEIGHTS = "model-kernel-epoch=47-epoch_val_loss=0.03.ckpt"
DEFAULT_NUMBER_CONFS = 200
DEFAULT_BATCH_SIZE = 100
DEFAULT_STEPS = 25
DEFAULT_MDS_NUM_INIT = 4
DEFAULT_STRUCTURE_GEN = "mds"
CONVERT_ANGSTROM_TO_NM = 10
MAX_SEQUENCE_LENGTH = 384  # set longest sequence the model can work on


# model model-kernel-epoch=47-epoch_val_loss=0.03.ckpt has  a UNET_LABELS_DIM of 512
# model model-kernel-epoch=47-epoch_val_loss=0.03.ckpt has a UNET_LABELS_DIM of 384
UNET_LABELS_DIM = 512

# Path to user config file
USER_CONFIG_PATH = os.path.expanduser(
    os.path.join("~/", ".starling_weights", "configs.py")
)


##
## The code block below lets us over-ride default values based on the configs.py file in the
## ~/.starling_weights directory
##


def load_user_config():
    """Load user configuration if the file exists and override default values."""
    if os.path.exists(USER_CONFIG_PATH):
        spec = importlib.util.spec_from_file_location("user_config", USER_CONFIG_PATH)
        user_config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(user_config)

        for key, value in vars(user_config).items():
            if not key.startswith("__") and key in globals():
                old_value = globals()[key]
                globals()[key] = value
                print(f"[Starling Config] Overriding {key}: {old_value} → {value}")


# Load user-defined config if available
load_user_config()

### Derived default values

# default paths to the model weights
DEFAULT_ENCODER_WEIGHTS_PATH = fix_ref_to_home(
    os.path.join(DEFAULT_MODEL_DIR, DEFAULT_ENCODE_WEIGHTS)
)
DEFAULT_DDPM_WEIGHTS_PATH = fix_ref_to_home(
    os.path.join(DEFAULT_MODEL_DIR, DEFAULT_DDPM_WEIGHTS)
)

# Github Releases URLs for model weights
GITHUB_ENCODER_URL = "https://github.com/idptools/starling/releases/download/v1.0.0/model-kernel-epoch.99-epoch_val_loss.1.72.ckpt"
GITHUB_DDPM_URL = "https://github.com/idptools/starling/releases/download/v1.0.0/model-kernel-epoch.47-epoch_val_loss.0.03.ckpt"

# Update default paths to check Hub first
DEFAULT_ENCODER_WEIGHTS_PATH = os.environ.get(
    "STARLING_ENCODER_PATH", GITHUB_ENCODER_URL
)
DEFAULT_DDPM_WEIGHTS_PATH = os.environ.get("STARLING_DDPM_PATH", GITHUB_DDPM_URL)

# Set the default number of CPUs to use
DEFAULT_CPU_COUNT_MDS = min(DEFAULT_MDS_NUM_INIT, os.cpu_count())

# define valid amino acids
VALID_AA = "ACDEFGHIKLMNPQRSTVWY"

# define conversion dictionaries for AAs
AA_THREE_TO_ONE = {
    "ALA": "A",
    "CYS": "C",
    "ASP": "D",
    "GLU": "E",
    "PHE": "F",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LYS": "K",
    "LEU": "L",
    "MET": "M",
    "ASN": "N",
    "PRO": "P",
    "GLN": "Q",
    "ARG": "R",
    "SER": "S",
    "THR": "T",
    "VAL": "V",
    "TRP": "W",
    "TYR": "Y",
}

AA_ONE_TO_THREE = {}
for x in AA_THREE_TO_ONE:
    AA_ONE_TO_THREE[AA_THREE_TO_ONE[x]] = x



import torch
from einops import rearrange
from torch import nn
from torch.nn import functional as F

from starling.models.normalization import RMSNorm


class CrossAttention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        context_dim: int,
        custom,
        channel_last=False,
    ) -> None:
        """
        CrossAttention module for use in UNet models. This module is used to
        perform attention on query (distance maps/2D data) using key and value
        (sequence labels). It is used to attend to 2D features using text/sequence
        embeddings, effectively conditioning the model on the text/sequence labels.

        Parameters
        ----------
        embed_dim : int
            Dimension of the input embedding
        num_heads : int
            Number of heads for multi-head attention
        kernel_size : int, optional
            Size of the kernel for generating query, by default 1
        """
        super(CrossAttention, self).__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.channel_last = channel_last
        self.context_dim = context_dim
        self.custom = custom

        assert (
            self.head_dim * num_heads == embed_dim
        ), "embed_dim must be divisible by num_heads"

        self.query_norm = nn.LayerNorm(embed_dim)
        self.context_norm = nn.LayerNorm(context_dim)

        self.query_proj = nn.Linear(embed_dim, embed_dim, bias=False)
        self.key_proj = nn.Linear(context_dim, embed_dim, bias=False)
        self.value_proj = nn.Linear(context_dim, embed_dim, bias=False)

        self.out_proj = nn.Linear(embed_dim, embed_dim)

    def forward(self, query, context=None):
        batch_size, channels, height, width = query.size()

        if not self.channel_last:
            query = rearrange(query, "b c h w -> b h w c")

        # Prenormalization of the query and context
        query = self.query_norm(query)
        context = self.context_norm(context)

        # Linear projection for the query (image features)
        Q = self.query_proj(query)  # [batch_size, height, width, channels]

        # Linear projections for the key and value (text embeddings)
        K = self.key_proj(context)  # [batch_size, seq_len, head_dim * num_heads]
        V = self.value_proj(context)  # [batch_size, seq_len, head_dim * num_heads]

        # Reshape query (image features) to match multi-head attention dimensions
        Q = rearrange(Q, "b x y (h d) -> b h (x y) d", h=self.num_heads)

        # Reshape key and value (text embeddings) for multi-head attention
        K = rearrange(K, "b s (h d) -> b h s d", h=self.num_heads)
        V = rearrange(V, "b s (h d) -> b h s d", h=self.num_heads)

        # Use scaled_dot_product_attention
        if self.custom:
            scores = torch.matmul(Q, K.transpose(-2, -1)) / (self.head_dim**0.5)
            attention_weights = F.softmax(scores, dim=-1)
            attention_output = torch.matmul(attention_weights, V)
        else:
            attention_output = F.scaled_dot_product_attention(Q, K, V)

        # Reshape back to original dimensions
        attention_output = rearrange(
            attention_output, "b h (x y) d -> b x y (h d)", x=height, y=width
        )
        attention_output = self.out_proj(attention_output)

        if not self.channel_last:
            attention_output = rearrange(attention_output, "b h w c -> b c h w")

        return attention_output


class SelfAttention(nn.Module):
    def __init__(
        self, embed_dim: int, num_heads: int, custom, channels_last: bool = False
    ) -> None:
        """
        This is a basic self-attention module. It uses linear layers to project
        the input into query, key, and value matrices, then performs scaled dot-product
        attention on these matrices. The output is then projected back to the original
        embedding dimension. Commonly used in transformer models.

        Parameters
        ----------
        embed_dim : int
            Dimension of the input embedding
        num_heads : int
            Number of heads for multi-head attention
        channels_last : bool, optional
            Whether the input has channels last format, if not it will be rearranged, by default False
        """
        super(SelfAttention, self).__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.channels_last = channels_last
        self.custom = custom

        assert (
            self.head_dim * num_heads == embed_dim
        ), "embed_dim must be divisible by num_heads"

        self.query_norm = nn.LayerNorm(embed_dim)

        self.query_proj = nn.Linear(embed_dim, embed_dim, bias=False)
        self.key_proj = nn.Linear(embed_dim, embed_dim, bias=False)
        self.value_proj = nn.Linear(embed_dim, embed_dim, bias=False)

        self.out_proj = nn.Linear(embed_dim, embed_dim)

    def forward(self, x):
        input_dim = x.dim()

        if input_dim == 4:
            batch_size, channels, height, width = x.size()
        elif input_dim == 3:
            batch_size, seq_len, channels = x.size()
        else:
            raise ValueError("Input dimension not supported")

        if not self.channels_last and input_dim == 4:
            x = rearrange(x, "b c h w -> b h w c")

        # Prenormalization
        x = self.query_norm(x)

        # Linear projection for the query
        Q = self.query_proj(x)
        K = self.key_proj(x)
        V = self.value_proj(x)

        if input_dim == 4:
            # If input is 4D (images)
            Q = rearrange(Q, "b x y (h d) -> b h (x y) d", h=self.num_heads)
            K = rearrange(K, "b x y (h d) -> b h (x y) d", h=self.num_heads)
            V = rearrange(V, "b x y (h d) -> b h (x y) d", h=self.num_heads)
        elif input_dim == 3:
            # If input is 3D (text)
            Q = rearrange(Q, "b x (h d) -> b h x d", h=self.num_heads)
            K = rearrange(K, "b x (h d) -> b h x d", h=self.num_heads)
            V = rearrange(V, "b x (h d) -> b h x d", h=self.num_heads)

        # Scaled Dot-Product Attention
        if self.custom:
            scores = torch.matmul(Q, K.transpose(-2, -1)) / (self.head_dim**0.5)
            attention_weights = F.softmax(scores, dim=-1)
            attention_output = torch.matmul(attention_weights, V)
        else:
            attention_output = F.scaled_dot_product_attention(Q, K, V)

        # Concatenate heads and reshape back to original dimensions
        if input_dim == 4:
            attention_output = rearrange(
                attention_output, "b h (x y) d -> b x y (h d)", x=height, y=width
            )
        elif input_dim == 3:
            attention_output = rearrange(
                attention_output, "b h x d -> b x (h d)", x=seq_len
            )

        attention_output = self.out_proj(attention_output)

        if not self.channels_last and input_dim == 4:
            attention_output = rearrange(attention_output, "b h w c -> b c h w")

        return attention_output


# The attention pooling could be used as an additional conditioning mechanism where its concatenated with
# timestep embeddings and then added to ResNet blocks (either in the middle or at the beginning)
# - Imagen seems to this at the beginning of the ResNet blocks
class AttentionPooling(nn.Module):
    def __init__(self, feature_dim, hidden_dim):
        super(AttentionPooling, self).__init__()
        self.attention = nn.Sequential(
            nn.SiLU(),  # Swish activation function
            nn.Linear(feature_dim, hidden_dim),
            nn.Tanh(),
            nn.Linear(hidden_dim, 1),
        )

    def forward(self, x):
        # x: input features of shape (batch_size, num_features, feature_dim)
        batch_size, num_features, feature_dim = x.size()

        # Compute attention scores
        attention_scores = self.attention(x)  # shape: (batch_size, num_features, 1)
        attention_weights = torch.softmax(
            attention_scores, dim=1
        )  # shape: (batch_size, num_features, 1)

        # Compute weighted sum of features
        pooled_features = torch.sum(
            attention_weights * x, dim=1
        )  # shape: (batch_size, feature_dim)

        return pooled_features


class SelfAttentionConv(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, kernel_size: int = 1) -> None:
        """
        SelfAttentionConv module for use in UNet models. This module is used to
        perform self-attention on 2D data. It is used to attend to spatial features
        in the 2D data, effectively allowing the model to learn spatial relationships
        between pixels.

        Parameters
        ----------
        embed_dim : int
            Dimension of the input embedding
        num_heads : int
            Number of heads for multi-head attention
        kernel_size : int, optional
            Size of the kernel for generating query, key, and value matrices, by default 1
        """
        super(SelfAttentionConv, self).__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads

        assert (
            self.head_dim * num_heads == embed_dim
        ), "embed_dim must be divisible by num_heads"

        self.query_conv = nn.Conv2d(
            embed_dim, embed_dim, kernel_size=kernel_size, padding=kernel_size // 2
        )
        self.key_conv = nn.Conv2d(
            embed_dim, embed_dim, kernel_size=kernel_size, padding=kernel_size // 2
        )
        self.value_conv = nn.Conv2d(
            embed_dim, embed_dim, kernel_size=kernel_size, padding=kernel_size // 2
        )
        self.out_conv = nn.Sequential(
            nn.Conv2d(embed_dim, embed_dim, kernel_size=1), RMSNorm(embed_dim)
        )

    def forward(self, x: torch.Tensor):
        batch_size, channels, height, width = x.size()

        # Convolutional projections
        Q = self.query_conv(x)
        K = self.key_conv(x)
        V = self.value_conv(x)

        # Reshape to (batch_size, num_heads, head_dim, height * width)
        Q = Q.view(batch_size, self.num_heads, self.head_dim, -1)
        K = K.view(batch_size, self.num_heads, self.head_dim, -1)
        V = V.view(batch_size, self.num_heads, self.head_dim, -1)

        # Transpose for multi-head attention (batch_size, num_heads, height * width, head_dim)
        Q = Q.transpose(2, 3)
        K = K.transpose(2, 3)
        V = V.transpose(2, 3)

        # Scaled Dot-Product Attention

        scores = torch.matmul(Q, K.transpose(-2, -1)) / (self.head_dim**0.5)
        attention_weights = F.softmax(scores, dim=-1)
        attention_output = torch.matmul(attention_weights, V)

        # Concatenate heads and reshape back to original dimensions
        attention_output = (
            attention_output.transpose(2, 3)
            .contiguous()
            .view(batch_size, self.embed_dim, height, width)
        )
        attention_output = self.out_conv(attention_output)

        return attention_output
